Package: fdaSP
Type: Package
Title: Sparse Functional Data Analysis Methods
Version: 1.1.3
Date: 2026-07-21
Author: Mauro Bernardi [aut, cre],
  Marco Stefanucci [aut],
  Antonio Canale [ctb]
Authors@R: c(person("Mauro", "Bernardi", role= c("aut","cre"), email="mauro.bernardi@unipd.it"),
            person("Marco", "Stefanucci", role= c("aut"), email="marco.stefanucci@uniroma2.it"),
	    person("Antonio", "Canale", role= "ctb", email="antonio.canale@unipd.it"))
Maintainer: Mauro Bernardi <mauro.bernardi@unipd.it>
Description: Provides algorithms to fit linear regression models under several popular penalization techniques and functional linear regression models based on Majorizing-Minimizing (MM) and Alternating Direction Method of Multipliers (ADMM) techniques.
    See Boyd et al (2010) <doi:10.1561/2200000016> for complete introduction to the method.
License: GPL (>= 3)
Suggests: RColorBrewer, gglasso, glmnet, latex2exp, utils
Imports: Rcpp, Rdpack, grDevices, graphics, stats, parallel,
        doParallel, foreach, splines, ks, pracma, CVXR
RdMacros: Rdpack
LinkingTo: Rcpp, RcppArmadillo
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-07-22 17:08:26 UTC; maurobernardi
Repository: CRAN
Date/Publication: 2026-07-22 17:50:21 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-07-22 18:42:55 UTC; unix
Archs: fdaSP.so.dSYM
