useDynLib(fdaSP, .registration=TRUE)
importFrom("Rcpp", "evalCpp")
importFrom("Rdpack","reprompt")
importFrom("grDevices", "rgb")
importFrom("graphics", "polygon")
importFrom("stats", "acf", "ccf", "lag", "na.omit", "pacf", "pnorm",
           "qnorm", "sd", "uniroot", "var")
importFrom("parallel", "detectCores", "makeCluster")
importFrom("doParallel", "registerDoParallel")
importFrom("foreach", "%dopar%")
importFrom("splines", "bs")
importFrom("ks", "invvec", "vec")
importFrom("pracma", "strcmpi")
importFrom("CVXR", "Variable", "norm2", "Maximize", "Problem", "psolve")

## export all regularly named functions
## (but allow for private functions whose name starts with a dot).name <- function(...)
##exportPattern("^[[:alpha:]]+")
export(lmSP)
export(lmSP_cv)
export(f2sSP)
export(f2sSP_cv)
export(f2fSP)
export(f2fSP_cv)
export(softhresh)
export(confband)
export(forward_diff)