\name{fdaSP-package}
\alias{fdaSP-package}
\alias{fdaSP}
\docType{package}
\title{\packageTitle{fdaSP}}
\description{\packageDescription{fdaSP}}
\section{Package Content}{\packageIndices{fdaSP}}
\author{\packageAuthor{fdaSP}}
\section{Maintainer}{\packageMaintainer{fdaSP}}
\keyword{package}
